'''
This is the reloading script of an ANN already trained 
Version 02 04/11/2021
'''

import pandas as pd
from sklearn.model_selection import train_test_split
from tensorflow.keras.utils import plot_model
from tensorflow.keras.layers import Dropout
from tensorflow.keras.optimizers import SGD
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras.callbacks import ModelCheckpoint
import matplotlib.pyplot as plt
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input
from tensorflow.keras.layers import Dense
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.optimizers import SGD
from keras.initializers import RandomUniform, Initializer, Constant
from statsmodels.graphics.gofplots import qqplot
import time
import numpy as np
import scipy as sp
from scipy.stats import zscore
from keras.models import load_model

#############################################################################################################
'''
Here starts the definition of a keras custom layer based on gaussian function
More info can be found on https://www.kaggle.com/residentmario/radial-basis-networks-and-custom-keras-layers
'''
from keras.layers import Layer
from keras import backend as K
K.clear_session()

class RBFLayer(Layer):
    def __init__(self, output_dim, initializer=None, betas=1.0, **kwargs):
        self.output_dim = output_dim
        self.init_betas = betas
        self.initializer = RandomUniform(0.0, 1.0)
        super(RBFLayer, self).__init__(**kwargs)

    def build(self, input_shape):
#         print(input_shape)
#         print(self.units)
        self.centers = self.add_weight(name='centers',
                                       shape=(self.output_dim, input_shape[1]),
                                       initializer=self.initializer,
                                       trainable=True)
        self.betas = self.add_weight(name='betas',
                                     shape=(self.output_dim,),
                                     initializer=Constant(
                                         value=self.init_betas),
                                     # initializer='ones',
                                     trainable=True)

        super(RBFLayer, self).build(input_shape)

    def call(self, inputs):
        C = K.expand_dims(self.centers)
        H = K.transpose(C-K.transpose(inputs))
        return K.exp(-self.betas * K.sum(H**2, axis=1))

    def compute_output_shape(self, input_shape):
        return (input_shape[0], self.output_dim)
    def get_config(self):
        # have to define get_config to be able to use model_from_json
        config = {
            'output_dim': self.output_dim
        }
        base_config = super(RBFLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))
    @classmethod
    def from_config(cls, config):
        return cls(**config)
#Here ends the definition of the Keras custom layer
###########################################################################################################

#

'''
The following instruction permits to fully load the already trained model with its weights
The h5 file contains:
the architecture of the model, allowing to re-create the model;
the weights of the model;
the training configuration (loss, optimizer);
the state of the optimizer. allowing to resume training exactly where you left off

'''
from sklearn.metrics import r2_score
    
loaded_model = load_model('best_dish_model_archive.h5', custom_objects={'RBFLayer': RBFLayer})
loaded_model.summary()

# make a single prediction on 12 parameters
row_input = [119.0,729.0,756.0,205.0666667,20.24333333,1.79,1.981666667,127.85,66.78333333,1003.0,108.55695,49.38975]
prediction = loaded_model.predict([row_input])
# print the prediction
print('Predicted: %.3f' % prediction)

# print the description of the network stored in the h5 file
print(loaded_model.summary())
# make vector predictions on 12 parameters 


'''
X_test.txt is the file in which there is a collction of measured input variables records.
X_test is the input dataframe created from X_test.txt. It has not been not used for training
prediction is the output dataframe calculated by the ANN using as input X_test
y_test.txt  is the file in which there is a collection of measuered output records.
'''


X = pd.read_csv('X_test.txt', sep=',', header=0) 
X_test=X.filter(regex='clean_day|_Direct_Normal_Irradiance|_Global_Horizontal_Irradiance|_Diffuse_Horizontal_Irradiance|_Ambient_Temperature|_Average_Wind_Speed|_Wind_Speed|_Wind_Direction|_Humidity|_Air_Pressure|_Solar_azimuth_position|_Solar_elev_position')
xx = X_test.astype('float64')
prediction = list(loaded_model.predict(xx))
prediction = pd.DataFrame.from_records(prediction)
prediction.columns = ['power_forecast']
column1vv=pd.read_csv('y_test.txt', sep=',', header=0)
column1v=column1vv['_Total_CSP_net_output_power']
R=r2_score(column1v, prediction)
print(R)

fig, ax = plt.subplots()
ax.scatter(column1v, prediction)
ax.plot([column1v.min(), column1v.max()], [column1v.min(), column1v.max()], 'k--', lw=4)
ax.set_xlabel('Measured')
ax.set_ylabel('Predicted')

ax.set_title('R²: ' + str(r2_score(column1v, prediction)))

# Saving figure by changing parameter values
plt.savefig("R².jpeg", facecolor='w', bbox_inches="tight", pad_inches=0.3, transparent=True)
plt.show()


'''
definition of graphs and charts
'''

# residuals definition

prediction.columns = ['power']
column1p=prediction['power']


residual1 = column1v-column1p
print(residual1.describe())  
import os
file_path = os.path.realpath(__file__)
dir_path = os.path.dirname(file_path)
print('MAE=')
print((residual1.abs()).mean(axis=0))

# store residuals and previsions in txt files 
residual1.to_csv(dir_path+'/residuals.txt')
prediction.to_csv(dir_path+'/previsions.txt')

# Plot the histogram distribution of residuals compared with an ideal normal distribution. 
def HISTOGRAM_plot():
    mu=residual1.mean()
    sigma=pd.DataFrame.std(residual1)
    num_bins = 100
    fig, ax = plt.subplots()
    # the histogram of the data
    n, bins, patches = ax.hist(residual1, num_bins, density=1)
    # add a 'best fit' line
    y = ((1 / (np.sqrt(2 * np.pi) * sigma))*np.exp(-0.5 * (1 / sigma * (bins - mu))**2))
    ax.plot(bins, y, '--')
    ax.set_xlabel('Power [W]')
    ax.set_ylabel('Probability density')
    ax.set_title(r'Histogram of residuals: $\mu=$'+str(mu)+', $\sigma=$'+str(sigma))
    # Tweak spacing to prevent clipping of ylabel
    fig.tight_layout()
    # Saving figure by changing parameter values
    plt.savefig("histogram_plot.jpeg", facecolor='w', bbox_inches="tight", pad_inches=0.3, transparent=True)
    return plt.show()

# A Q-Q plot, or quantile plot, compares two distributions and can be used to see #how similar or different they happen to be.
# We can create a Q-Q plot using the qqplot() function in the statsmodels library. # The Q-Q plot can be used to quickly check
# the normality of the distribution of residual errors.
def QQ_plot():
    qqplot(residual1)
    return plt.show()
def QQ_plot2():
    fig, ax = plt.subplots(figsize=(6,2.5))
    _, (__, ___, r) = sp.stats.probplot(residual1, plot=ax, fit=True)
    r**2   
    # Saving figure by changing parameter values
    plt.savefig("QQ_plot.jpeg", facecolor='w', bbox_inches="tight", pad_inches=0.3, transparent=True)
    return plt.show()

# Plot the predicted value against the expected measured value
def PREvsMEA():
    fig, ax = plt.subplots()
    ax.scatter(column1v, column1p)
    ax.plot([column1v.min(), column1v.max()], [column1v.min(), column1v.max()], 'k--', lw=4)
    ax.set_xlabel('Measured')
    ax.set_ylabel('Predicted')
    ax.set_title('R²: ' + str(r2_score(column1v, prediction)))
    # Saving figure by changing parameter values
    plt.savefig("p-vs-m_plot.jpeg", facecolor='w', bbox_inches="tight", pad_inches=0.3, transparent=True)
    return plt.show()

# Call all the routines to generate the images
HISTOGRAM_plot()
QQ_plot2()
PREvsMEA()

# Destroys and resets all variables to avoid problems running two ANN
import sys
sys.modules[__name__].__dict__.clear()    # clear all 


