'''
This is the ANN definition script
please note that the original dataset of experimental data is not supplied
only a validation subset is given, in order to reload the ANN from h5 file 
'''



import pandas as pd
import seaborn as sns
from sklearn.model_selection import train_test_split
from tensorflow.keras.utils import plot_model
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras.callbacks import ModelCheckpoint
import matplotlib.pyplot as plt
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input
from tensorflow.keras.layers import Dense
from tensorflow.keras.optimizers import Adam, SGD, Adamax, Ftrl, Nadam, RMSprop
from keras.initializers import RandomUniform, Initializer, Constant
from statsmodels.graphics.gofplots import qqplot
import time
import numpy as np
import scipy as sp
from scipy.stats import zscore
import os
from sklearn.metrics import r2_score
file_path = os.path.realpath(__file__)
dir_path = os.path.dirname(file_path)


'''
Here starts the definition of a keras custom layer based on gaussian function
More info can be found on https://www.kaggle.com/residentmario/radial-basis-networks-and-custom-keras-layers
'''
from keras.layers import Layer
from keras import backend as K
K.clear_session()

class RBFLayer(Layer):
    def __init__(self, output_dim, initializer=None, betas=1.0, **kwargs):
        self.output_dim = output_dim
        self.init_betas = betas
        self.initializer = RandomUniform(0.0, 1.0)
        super(RBFLayer, self).__init__(**kwargs)

    def build(self, input_shape):
        self.centers = self.add_weight(name='centers',
                                       shape=(self.output_dim, input_shape[1]),
                                       initializer=self.initializer,
                                       trainable=True)
        self.betas = self.add_weight(name='betas',
                                     shape=(self.output_dim,),
                                     initializer=Constant(
                                         value=self.init_betas),
                                     # initializer='ones',
                                     trainable=True)

        super(RBFLayer, self).build(input_shape)

    def call(self, inputs):
        C = K.expand_dims(self.centers)
        H = K.transpose(C-K.transpose(inputs))
        return K.exp(-self.betas * K.sum(H**2, axis=1))

    def compute_output_shape(self, input_shape):
        return (input_shape[0], self.output_dim)
    def get_config(self):
        # have to define get_config to be able to use model_from_json
        config = {
            'output_dim': self.output_dim
        }
        base_config = super(RBFLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))
    @classmethod
    def from_config(cls, config):
        return cls(**config)
'''
Here ends the definition of a keras custom RBF layer based on gaussian function
'''


#reads the original dataset and put into a pandas dataframe; fields separator is tabulation  
original_df = pd.read_csv('dish_original_data.txt', sep='\t', header=0) 

#outliers removal
'''
Here starts a mathematical procedure for removing outliers
Given mu and sigma, a simple way to identify outliers is to compute a z-score
for every xi, which is defined as the number of standard deviations away xi
is from the mean […] Data values that have a z-score sigma greater than a
threshold, for example, of three, are declared to be outliers.
Following procedure apply removal on only 3 parameters:
DNI, air temperature and produced power energy
with the same threshold = thrsld
'''
thrsld = 2
###############################################################################################################################################################################################################################################################################################################################
z_scores = sp.stats.zscore(original_df['_Total_CSP_net_output_power'])
#calculates z-scores of dataframe
abs_z_scores = np.abs(z_scores)
clean_entries = (abs_z_scores < thrsld)
clean_df = original_df[clean_entries]

z_scores2 = sp.stats.zscore(clean_df['_Direct_Normal_Irradiance'])
#calculates z-scores of dataframe
abs_z_scores2 = np.abs(z_scores2)
clean_entries2 = (abs_z_scores2 < thrsld)
clean_df2 = clean_df[clean_entries2]

z_scores3 = sp.stats.zscore(clean_df2['_Ambient_Temperature'])
#calculates z-scores of dataframe
abs_z_scores3 = np.abs(z_scores3)
clean_entries3 = (abs_z_scores3 < thrsld)
clean_df3 = clean_df2[clean_entries3]

df_no_outliers=clean_df3.filter(regex='clean_day|_Direct_Normal_Irradiance|_Global_Horizontal_Irradiance|_Diffuse_Horizontal_Irradiance|_Ambient_Temperature|_Average_Wind_Speed|_Wind_Speed|_Wind_Direction|_Humidity|_Air_Pressure|_Solar_azimuth_position|_Solar_elev_position|_Total_CSP_net_output_power')
dfinput_no_outliers=clean_df3.filter(regex='clean_day|_Direct_Normal_Irradiance|_Global_Horizontal_Irradiance|_Diffuse_Horizontal_Irradiance|_Ambient_Temperature|_Average_Wind_Speed|_Wind_Speed|_Wind_Direction|_Humidity|_Air_Pressure|_Solar_azimuth_position|_Solar_elev_position')
dfoutput_no_outliers=clean_df3.filter(regex='_Total_CSP_net_output_power')
###############################################################################################################################################################################################################################################################################################################################
y = dfoutput_no_outliers.astype('float64')
X = dfinput_no_outliers.astype('float64')

#definition of input and output validation and training datasesets.
#in this case validation is only 15% of the global data without outliers
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.15)
print(X_train.shape, X_test.shape, y_train.shape, y_test.shape)

#saving of input and output validation and training set
X_train.to_csv(dir_path+'/X_train.txt')
X_test.to_csv(dir_path+'/X_test.txt')
y_train.to_csv(dir_path+'/y_train.txt')
y_test.to_csv(dir_path+'/y_test.txt')


n_features = X_train.shape[1]
# define model architecture
init = RandomUniform(minval=0, maxval=1)



n_inputs, n_outputs = X.shape[1], y.shape[1]
################################################################################################################
# input layer
visible = Input(shape=(12,))
dd = Dense(12, activation='relu', kernel_initializer='he_uniform')(visible)
dd = BatchNormalization()(dd)
layer1 = Dense(130, activation='relu', kernel_initializer='he_uniform')(dd)
layer2 = Dropout(0.003)(layer1)
layer3 = Dense(200, activation='relu', kernel_initializer='he_uniform')(layer2)
#############################################
# this is the call of a RBF layer	     #
#layer4 = RBFLayer(130, betas=0.5)(layer3)  #
#############################################
layer4 = Dense(4, activation='relu', kernel_initializer='he_uniform')(layer3)
layer5 = Dense(700, activation='relu', kernel_initializer='he_uniform')(layer4)
layer6 = Dense(100, activation='relu', kernel_initializer='he_uniform')(layer5)
layer7 = Dense(50, activation='relu', kernel_initializer='he_uniform')(layer6)
output1 = Dense(1, activation='relu', kernel_initializer='he_uniform')(layer7)
###############################################################################################################
model = Model(inputs=visible, outputs=[output1])
# print an explicative image of the ANN
plot_model(model, to_file=dir_path+'/dish_model.png')

init_lr = 1e-4
# number of input sample used for a single epoch training.
# it is possible to use one sample or the complete lenght of Inputs_train dataframe
# this parameter influences the stability of the error gradient, used quantity of RAM and speed of training 
epoch_number=150000
batch=70
epochs = epoch_number

#different optimizer configurations that can be used in model construction
opt1 = Adam(lr=init_lr, decay=init_lr / epochs)
opt2 = Adam(learning_rate=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-07)
opt3 = SGD(lr=1e-2, momentum=0.9, decay=1e-7/epochs)
opt4 = Adamax(learning_rate=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-07)
opt5 = Ftrl(learning_rate=0.001, learning_rate_power=-0.5, initial_accumulator_value=0.1, l1_regularization_strength=0.0, l2_regularization_strength=0.0, l2_shrinkage_regularization_strength=0.0, beta=0.0,)
opt6 = Nadam(learning_rate=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-07)
opt7 = RMSprop(learning_rate=0.001, rho=0.9, momentum=0.0, epsilon=1e-07, centered=False)

#model construction
model.compile(opt2, loss='mae', metrics='mae')
# definition of the file where the net weights will be stored. 
# The file can then be used to make instant predictions without the need for further training
archive = dir_path+'/dish_model_archive.h5'

# checkpoint configuration that can be used to avoid overfitting monitoring 
# this control prevents the network from training very well on training data at the expense of performance on validation data.
# val_loss is the value of cost function for your validation data.

checkpoint3 = ModelCheckpoint(archive, monitor='val_loss', verbose=1, save_weights_only=False, save_best_only=True, mode='auto')

# configure early stopping of training that can be used to avoid overfitting
# improvement of val_loss are considered only if differences is > min_delta
# a sequence of epochs = patience without improvements will stop training
es1 = EarlyStopping(monitor='val_loss', min_delta=1e-2, patience=70)
es2 = EarlyStopping(monitor='val_loss', min_delta=1e-2, patience=100)
es3 = EarlyStopping(monitor='val_loss', min_delta=1e-3, patience=70, restore_best_weights=True)
es4 = EarlyStopping(monitor='val_loss', min_delta=0.5*1e-4, patience=500, restore_best_weights=True)
es5 = EarlyStopping(monitor='val_loss', min_delta=0.1*1e-3, patience=5000, restore_best_weights=True)

#definition of callback to avoid overfitting and stop training 
callbacks_list = [checkpoint3,es5]

#to disable early stopping  callbacks_list =[]

#this file is used to check if a training is running
f = open(dir_path+'/lockfile', 'w')
f.close()

# this variable is uused to measure the training time
start = time.perf_counter()

# in the history variable is stored the performance of the model during training phase
history=model.fit(X_train, y_train, verbose=0,epochs=epoch_number, batch_size=batch, callbacks=callbacks_list, validation_data=(X_test, y_test))

# when the training phase is ended, the file locking is deleted. 
# useful to quick test if ANN is running or if there was an unexpected error
os.remove(dir_path+'/lockfile')

# Calculates elapsed time
elapsed = time.perf_counter() - start

# print in the terminal the elapsed time in seconds
print('%.3f seconds have elapsed' % elapsed)

# print in the terminal the elapsed epochs
n_epochs = len(history.history['val_loss']) 

# calculates the training speed of the ANN in terms of seconds per epoch.
speed=float(elapsed)/float(n_epochs)

# print in the terminal the training speed
print('Velocity is  %.3f seconds per epoch.' % speed)

# Store in a txt file the info about elapsed file and training speed
lines = ['%.3f seconds have elapsed' % elapsed, 'Velocity is  %.3f seconds per epoch.' % speed]
with open('readme.txt', 'w') as f2:
    f2.write('\n'.join(lines))
f2.close()


# create a textual summary of your deep learning model and print it into a terminal.
# it is possible to see the output shape and number of weights in each layer.

print(model.summary())

# evaluate model on all validation dataset computing the mean absolute error between measures and predictions.
mae = model.evaluate(X_test, y_test, verbose=0)

# save model and architecture to single file useful to predict without further training
model.save(dir_path+'/dish_model_archive.h5')
print("Saving model to disk")

# print in a terminal the value of mae
print('>%.3f' % mae[0])


# the next section defines some variables useful for calculating residuals
# residual is the difference between the observed value and the estimated value of the quantity of interest 
# calculation of residuals is useful to assess the performance of the trained ANN
prediction = list(model.predict(X_test))
prediction = pd.DataFrame.from_records(prediction)

# assign a column name
prediction.columns = ['power']
column1p=prediction['power']
column1v=y_test['_Total_CSP_net_output_power']
column1v.reset_index(drop=True,inplace=True)
residual1 = column1v-column1p

#The next commands make a statistical analysis of residuals
print(residual1.describe())  
stats = df_no_outliers.describe()
df_stats = stats.transpose()

# save model data from dataframes to csv
residual1.to_csv(dir_path+'/residuals.txt')
history_df = pd.DataFrame.from_records(history.history)
history_df.to_csv(dir_path+'/history.txt')
residual1.to_csv(dir_path+'/residuals.txt')
prediction.to_csv(dir_path+'/previsions.txt')
clean_df3.to_csv(dir_path+'/no_outliers_data.txt')
df_stats.to_csv(dir_path+'/data_stats.txt')


#The next commands make a statistical analysis of residuals
def Analysis_plot():
    sns.pairplot(df_no_outliers, diag_kind="gnome")
    plt.savefig("Analysis_plot.jpeg", facecolor='w', bbox_inches="tight", pad_inches=0.3, transparent=True)
    return #plt.show()

#Check the relation between the variables using a correlation graph.
def Correlation_plot():
    df_corr=df_no_outliers.corr()
    fig, ax = plt.subplots(figsize=(12, 10))
    mask = np.triu(np.ones_like(df_corr, dtype=bool))
    mask = mask[1:, :-1]
    corr = df_corr.iloc[1:,:-1].copy()
    cmap = sns.diverging_palette(0, 230, 90, 60, as_cmap=True)
    sns.heatmap(corr, mask=mask, annot=True, fmt=".2f", linewidths=5, cmap=cmap, vmin=-1, vmax=1, cbar_kws={"shrink": .8}, square=True)
    yticks = [i.upper() for i in corr.index]
    xticks = [i.upper() for i in corr.columns]
    plt.yticks(plt.yticks()[0], labels=yticks, rotation=0)
    plt.xticks(plt.xticks()[0], labels=xticks)
    title = 'Correlation matrix'
    plt.title(title, loc='left', fontsize=18)
    plt.savefig("Correlation_plot.jpeg", facecolor='w', bbox_inches="tight", pad_inches=0.3, transparent=True)
    return plt.show()

# Plot history: MAE on dish-Stirling power forecasting; It is the performance of ANN during training
def MAE_plot():
    plt.plot(history.history['loss'], label='MAE on training data')
    plt.plot(history.history['val_loss'], label='MAE on validation data')
    plt.title('MAE on dish-Stirling power forecasting')
    plt.ylabel('MAE [kW]')
    plt.xlabel('Epochs')
    plt.legend(loc="upper left")
    # Saving figure by changing parameter values
    plt.savefig("MAE_plot.jpeg", facecolor='w', bbox_inches="tight", pad_inches=0.3, transparent=True)
    return plt.show()

# Plot the histogram distribution of residuals compared with an ideal normal distribution. 
def HISTOGRAM_plot():
    mu=residual1.mean()
    sigma=pd.DataFrame.std(residual1)
    num_bins = 100
    fig, ax = plt.subplots()
    # the histogram of the data
    n, bins, patches = ax.hist(residual1, num_bins, density=1)
    # add a 'best fit' line
    y = ((1 / (np.sqrt(2 * np.pi) * sigma))*np.exp(-0.5 * (1 / sigma * (bins - mu))**2))
    ax.plot(bins, y, '--')
    ax.set_xlabel('Power [W]')
    ax.set_ylabel('Probability density')
    ax.set_title(r'Histogram of residuals: $\mu=$'+str(mu)+', $\sigma=$'+str(sigma))
    # Tweak spacing to prevent clipping of ylabel
    fig.tight_layout()
    # Saving figure by changing parameter values
    plt.savefig("histogram_plot.jpeg", facecolor='w', bbox_inches="tight", pad_inches=0.3, transparent=True)
    return plt.show()

# A Q-Q plot, or quantile plot, compares two distributions and can be used to see #how similar or different they happen to be.
# We can create a Q-Q plot using the qqplot() function in the statsmodels library. # The Q-Q plot can be used to quickly check
# the normality of the distribution of residual errors.
def QQ_plot():
    qqplot(residual1)
    return plt.show()
def QQ_plot2():
    fig, ax = plt.subplots(figsize=(6,2.5))
    _, (__, ___, r) = sp.stats.probplot(residual1, plot=ax, fit=True)
    r**2   
    # Saving figure by changing parameter values
    plt.savefig("QQ_plot.jpeg", facecolor='w', bbox_inches="tight", pad_inches=0.3, transparent=True)
    return plt.show()

# Plot the predicted value against the expected measured value
def PREvsMEA():
    fig, ax = plt.subplots()
    ax.scatter(column1v, column1p)
    ax.plot([column1v.min(), column1v.max()], [column1v.min(), column1v.max()], 'k--', lw=4)
    ax.set_xlabel('Measured')
    ax.set_ylabel('Predicted')
    ax.set_title('R²: ' + str(r2_score(column1v, prediction)))
    # Saving figure by changing parameter values
    plt.savefig("p-vs-m_plot.jpeg", facecolor='w', bbox_inches="tight", pad_inches=0.3, transparent=True)
    return plt.show()

# Call all the routines to generate the images
#Analysis_plot()
Correlation_plot()
MAE_plot()
HISTOGRAM_plot()
QQ_plot2()
PREvsMEA()
# Destroys and resets all variables to avoid problems running two ANN
import sys
sys.modules[__name__].__dict__.clear()    # clear all
